open Batteries
open Cfg
open Elang_print
open Prog
open Utils

let rec dump_cfgexpr : expr -> string = function
  | Ebinop(b, e1, e2) -> Format.sprintf "(%s %s %s)" (dump_cfgexpr e1) (dump_binop b) (dump_cfgexpr e2)
  | Eunop(u, e) -> Format.sprintf "(%s %s)" (dump_unop u) (dump_cfgexpr e)
  | Eint i -> Format.sprintf "%d" i
  | Evar s -> Format.sprintf "%s" s
  | Ecall(f,l) ->  let l_string_of_expr = List.map (fun e -> dump_cfgexpr e) l in
      f^"("^(String.concat "," l_string_of_expr)^")"
  | Estk(i) -> Format.sprintf "@%i" i
  | Eload(e,i) -> Format.sprintf "mem[@%s:+%i]" (dump_cfgexpr e) i

let dump_list_cfgexpr l =
  l |> List.map dump_cfgexpr |> String.concat ", "


let dump_arrows oc fname n (node: cfg_node) =
  match node with
  | Cassign (_, _, succ)
  | Cprint (_, succ)
  | Cnop succ ->
    Format.fprintf oc "n_%s_%d -> n_%s_%d\n" fname n fname succ
  | Creturn _ -> ()
  | Ccmp (_, succ1, succ2) ->
    Format.fprintf oc "n_%s_%d -> n_%s_%d [label=\"then\"]\n" fname n fname succ1;
    Format.fprintf oc "n_%s_%d -> n_%s_%d [label=\"else\"]\n" fname n fname succ2
  | Ccall(_,_,succ) ->  Format.fprintf oc "n_%s_%d -> n_%s_%d \n" fname n fname succ
  | Cstore(_,_,_,succ) ->  Format.fprintf oc "n_%s_%d -> n_%s_%d \n" fname n fname succ

  let dump_cfg_node oc (node: cfg_node) =
  match node with
  | Cassign (v, e, _) -> Format.fprintf oc "%s = %s" v (dump_cfgexpr e)
  | Cprint (e, _) -> Format.fprintf oc "print %s" (dump_cfgexpr e)
  | Creturn e -> Format.fprintf oc "return %s" (dump_cfgexpr e)
  | Ccmp (e, _, _) -> Format.fprintf oc "%s" (dump_cfgexpr e)
  | Cnop _ -> Format.fprintf oc "nop"
  | Ccall(fname,l_expr,succ) -> Format.fprintf oc "%s" (dump_cfgexpr (Ecall(fname,l_expr)))
  | Cstore(e_addr,e_val,sz,succ) -> Format.fprintf oc "mem[%s:+%i] = %s" (dump_cfgexpr e_addr) sz (dump_cfgexpr e_val) 


let dump_liveness_state oc ht state =
  Hashtbl.iter (fun n cn ->
      Format.fprintf oc "%a : " dump_cfg_node cn;
      let vs = Hashtbl.find_default state n Set.empty in
      Set.iter (fun v ->Format.fprintf oc "%s, " v) vs;
      Format.fprintf oc "\n";
      flush_all ()
    ) ht 

let dump_cfg_fun oc cfgfunname (f : cfg_fun) =
  Format.fprintf oc "subgraph cluster_%s {\n label=\"%s\";\n" cfgfunname cfgfunname;
  Hashtbl.iter (fun n node ->
      Format.fprintf oc "n_%s_%d [label=\"%a\" xlabel=\"%d\" shape=%s];\n" cfgfunname n dump_cfg_node node n (if n = f.cfgentry then "rectangle peripheries=2" else "rectangle");
      dump_arrows oc cfgfunname n node
    ) f.cfgfunbody;
  Format.fprintf oc "}\n"

let dump_cfg_prog oc (cp: cprog) =
  Format.fprintf oc "digraph G{\n";
  dump_prog dump_cfg_fun oc cp;
  Format.fprintf oc "\n}"

