open Ast
open Batteries
open Prog
open Utils

type binop = Eadd | Emul | Emod | Exor | Ediv | Esub (* binary operations *)
           | Eclt | Ecle | Ecgt | Ecge | Eceq | Ecne (* comparisons *)
type unop = Eneg

type expr =
  | Ebinop of binop * expr * expr
  | Eunop of unop * expr
  | Eint of int
  | Echar of char
  | Evar of string
  | Ecall of string * expr list
  | Eaddrof of expr
  | Eload of expr



type instr =
  | Iassign of string * expr
  | Iif of expr * instr * instr
  | Iwhile of expr * instr
  | Iblock of instr list
  | Ireturn of expr
  | Iprint of expr
  | Icall of string * expr list
  | Istore of expr * expr

type efun = {
  funargs: ( string * typ ) list;
  funbody: instr;
  funvartyp : (string,typ) Hashtbl.t;
  funrettype : typ;
  funvarinmem : (string, int) Hashtbl.t;
  funstksz : int;
}

type eprog = efun prog

let string_of_funvarinmem (funvarinmem : (string, int) Hashtbl.t) : string =
  Hashtbl.fold (fun name_var addr_var acc -> name_var^" : "^(Format.sprintf "%i" addr_var)^acc) funvarinmem ""