open Batteries
open BatList
open Elang
open Cfg
open Utils
open Prog

type reg = int

type rtl_arg = 
|Reg of reg
|Addr of int * int (*Addr(ofs,sz)*)

type rtl_cmp = Rclt | Rcle | Rcgt | Rcge | Rceq | Rcne

type rtl_instr = Rbinop of binop * reg * reg * reg
               | Runop of unop * reg * reg
               | Rconst of reg * int
               | Rbranch of rtl_cmp * reg * reg * int
               | Rjmp of int
               | Rmov of reg * reg
               | Rret of reg
               | Rlabel of int
               | Rprint of reg
               | Rcall of reg option * string * reg list
               | Rstk of reg * int
               | Rload of reg * reg * int
               | Rstore of reg * reg * int 

type rtl_fun = { rtlfunargs: rtl_arg list; (*liste des registres qui correspondes aux expressions qui sont les arguments :
 f(a) -> Rcall(None, f, registre représentant a / Addr memeoire représentant a) *)
                 rtlfunbody: (int, rtl_instr list) Hashtbl.t;
                 rtlfunentry: int;
                 rtlfuninfo: (string*reg) list;
                 rtlfunstksz : int
               }

let written_rtl_regs_instr (i: rtl_instr) =
  match i with
  | Rbinop (_, rd, _, _)
  | Runop (_, rd, _)
  | Rconst (rd, _)
  | Rmov (rd, _) -> Set.singleton rd
  | Rprint _
  | Rret _
  | Rlabel _
  | Rbranch (_, _, _, _)
  | Rjmp _ -> Set.empty
  | Rcall(ord,_,_) -> (match ord with 
                      |None -> Set.empty
                      |Some(rd) -> Set.singleton rd)
  | Rstk(rd,_) -> Set.singleton rd
  | Rload(rd,_,_) -> Set.singleton rd
  | Rstore(rd,_,_) -> Set.singleton rd
                 

let written_rtl_regs (l: rtl_instr list) =
  List.fold_left (fun acc i -> Set.union acc (written_rtl_regs_instr i))
    Set.empty l
